.ml4 {
    position: relative;
    font-weight: 900;
    font-size: 4.5em;    
    text-align: center;
  }
  .ml4 .letters {
    position: absolute;
    margin: auto;
    left: 0;
    top: -0.5em;
    right: 0;
    opacity: 0; 
  }