from moai.supervision.losses.regression.robust.berhu import Berhu
from moai.supervision.losses.regression.robust.welsch import Welsch
from moai.supervision.losses.regression.robust.wing import Wing
from moai.supervision.losses.regression.robust.gm import GemanMcClure

__all__ = [
    "Berhu",
    "Welsch",
    "Wing",
    "GemanMcClure",
]